from pathlib import Path

from nexom.web.path import Static
from nexom.web.request import Request
from nexom.core.error import PathNotFoundError

def test_static_serves_file(tmp_path: Path, make_environ):
    # tmp_path/static/hello.txt を作る
    static_dir = tmp_path / "static"
    static_dir.mkdir()
    (static_dir / "hello.txt").write_bytes(b"hi")

    s = Static("static/", str(static_dir), "Static")

    req = Request(make_environ(path="/static/hello.txt"))
    res = s.call_handler(req)

    assert res.body == b"hi"


def test_static_blocks_traversal(tmp_path: Path, make_environ):
    static_dir = tmp_path / "static"
    static_dir.mkdir()
    (tmp_path / "secret.txt").write_bytes(b"nope")

    s = Static("static/", str(static_dir), "Static")

    # ../secret.txt を狙う
    req = Request(make_environ(path="/static/../secret.txt"))
    try:
        s.call_handler(req)
        assert False, "should raise"
    except PathNotFoundError:
        assert True


import pytest

def test_middleware_chain_order(make_environ):
    """
    middleware 実装が入ってる場合だけ通るテスト。
    無い場合は skip。
    """
    try:
        from nexom.web.middleware import MiddlewareChain
    except Exception:
        pytest.skip("middleware module not available")

    from nexom.web.response import Response
    from nexom.web.request import Request

    events = []

    def mw1(req, args, next_):
        events.append("mw1-before")
        res = next_(req, args)
        events.append("mw1-after")
        return res

    def mw2(req, args, next_):
        events.append("mw2-before")
        res = next_(req, args)
        events.append("mw2-after")
        return res

    def handler(req, args):
        events.append("handler")
        return Response("ok")

    wrapped = MiddlewareChain((mw1, mw2)).wrap(handler)

    req = Request(make_environ(path="/"))
    res = wrapped(req, {})
    assert res.body == b"ok"

    assert events == [
        "mw1-before",
        "mw2-before",
        "handler",
        "mw2-after",
        "mw1-after",
    ]


# src/nexom/templates/auth.py
from __future__ import annotations

from importlib import resources
import pathlib as plb

from ..app.auth import AuthClient, KEY_NAME
from ..app.request import Request
from ..app.response import HtmlResponse, JsonResponse, Redirect, ErrorResponse
from ..app.cookie import Cookie
from ..app.path import Path, Router
from ..core.object_html_render import HTMLDoc, ObjectHTML
from ..core.error import NexomError, _status_for_auth_error


# --------------------
# Object HTML
# --------------------

_OHTML: ObjectHTML = ObjectHTML(
    HTMLDoc(
        "signup",
        resources.files("nexom.assets.auth_page").joinpath("signup.html").read_text(encoding="utf-8"),
    ),
    HTMLDoc(
        "login",
        resources.files("nexom.assets.auth_page").joinpath("login.html").read_text(encoding="utf-8"),
    ),
)


# --------------------
# Pages
# --------------------

class AuthPages(Path):
    def __init__(self,
        path: str,
        auth_server: str,
        *,
        login_path: str = "login/",
        signup_path: str = "signup/",
        logout_path: str = "logout/"
    ):
        self.auth_server = auth_server

        root = plb.Path(path)
        self.routing = Router(
            LoginPage(str(root / login_path), self.auth_server),
            SignupPage(str(root / signup_path), self.auth_server),
            LogoutPage(str(root / logout_path), self.auth_server),
        )

        super().__init__(path, self._handler, "UserAuthPage")

    def _handler(self, req: Request, args: dict) -> JsonResponse:

        p = self.routing.get(req.path)
        return p.call_handler(req)


class LoginPage(Path):
    def __init__(self, path: str, auth_server: str) -> None:
        super().__init__(path, self._handler, "LoginPage")

        self.client = AuthClient(auth_server)

    def _handler(self, req: Request, args: dict) -> JsonResponse:
        if req.method == "GET":
            return HtmlResponse(_OHTML.render("login", page_path=req.path))

        try:
            data = req.json() or {}
            sess = self.client.login(
                user_id=str(data.get("user_id") or ""),
                password=str(data.get("password") or ""),
            )

            set_cookie = Cookie(KEY_NAME, sess.token, Path="/", MaxAge=sess.expires_at)
            return JsonResponse({"ok": True, "user_id": sess.user_id, "token": sess.token, "expires_at": sess.expires_at}, cookie=str(set_cookie))
        except NexomError as e:
            return JsonResponse({"ok": False, "error": e.code}, status=_status_for_auth_error(e.code))

        except Exception:
            return JsonResponse({"ok": False, "error": "InternalError"}, status=500)
        
class SignupPage(Path):
    def __init__(self, path: str, auth_server: str) -> None:
        super().__init__(path, self._handler, "SignupPage")

        self.client = AuthClient(auth_server)

    def _handler(self, req: Request, args: dict) -> JsonResponse:
        if req.method == "GET":
            return HtmlResponse(_OHTML.render("signup", page_path=req.path))

        try:
            data = req.json() or {}
            self.client.signup(
                user_id=str(data.get("user_id") or ""),
                public_name=str(data.get("public_name") or ""),
                password=str(data.get("password") or ""),
            )
            return JsonResponse({"ok": True}, status=201)

        except NexomError as e:
            return JsonResponse({"ok": False, "error": e.code}, status=_status_for_auth_error(e.code))

        except Exception:
            return JsonResponse({"ok": False, "error": "InternalError"}, status=500)
        

class LogoutPage(Path):
    def __init__(self, path: str, auth_server: str) -> None:
        super().__init__(path, self._handler, "LogoutPage")
        self.client = AuthClient(auth_server)

    def _handler(self, req: Request, args: dict) -> Redirect | ErrorResponse:
        if req.method != "GET":
            return ErrorResponse(405, "Please make a GET request")

        token = req.cookie.get(KEY_NAME) if req.cookie else None

        redirect_url = req.headers.get("referer") or "/"
        if not redirect_url.startswith("/"):
            redirect_url = "/"

        if token:
            try:
                self.client.logout(token=token)
            except NexomError as e:
                return ErrorResponse(500, e.code)
            except Exception as e:
                return ErrorResponse(500, str(e))

        set_cookie = Cookie(KEY_NAME, "", Path="/", MaxAge=0)
        return Redirect(redirect_url, cookie=str(set_cookie))




from __future__ import annotations

import os
import re
from mimetypes import guess_type
from pathlib import Path as _Path
from typing import Optional

from ..core.error import (
    PathNotFoundError,
    PathlibTypeError,
    PathInvalidHandlerTypeError,
    PathHandlerMissingArgError,
)
from .request import Request
from .response import Response, JsonResponse
from .middleware import Middleware, MiddlewareChain, Handler


# ====================
# Path (base)
# ====================

class Path:
    """Represents a route with optional path arguments and its handler."""

    def __init__(
        self,
        path: str,
        handler: Handler,
        name: str,
        *,
        methods: set[str] | None = None,   # None = any method
    ):
        self.handler = handler
        self.name: str = name
        self.methods: set[str] | None = {m.upper() for m in methods} if methods else None

        path_segments = path.strip("/").split("/") if path.strip("/") else [""]
        self.path_args: dict[int, str] = {}

        detection_index = 0
        for idx, segment in enumerate(path_segments):
            m = re.match(r"{(.*?)}", segment)
            if m:
                if detection_index == 0:
                    detection_index = idx
                self.path_args[idx] = m.group(1)

        if detection_index == 0:
            detection_index = len(path_segments)

        self.path: str = "/".join(path_segments[:detection_index])
        self.detection_range: int = detection_index

    def _read_args(self, request_path: str) -> dict[str, Optional[str]]:
        """Build args for this request. (No shared state)"""
        args: dict[str, Optional[str]] = {}
        segments = request_path.strip("/").split("/") if request_path.strip("/") else [""]
        for idx, arg_name in self.path_args.items():
            args[arg_name] = segments[idx] if idx < len(segments) else None
        return args

    def call_handler(
        self,
        request: Request,
        middlewares: tuple[Middleware, ...] = (),
    ) -> Response:
        try:
            args = self._read_args(request.path)

            handler = self.handler
            if middlewares:
                handler = MiddlewareChain(middlewares).wrap(handler)

            res = handler(request, args)

            if isinstance(res, dict):
                return JsonResponse(res)

            if not isinstance(res, Response):
                raise PathInvalidHandlerTypeError(self.handler)

            return res

        except TypeError as e:
            # handler の引数不足
            if re.search(r"takes \d+ positional arguments? but \d+ were given", str(e)):
                raise PathHandlerMissingArgError()
            raise


# ====================
# Method specific paths
# ====================

class Get(Path):
    def __init__(self, path: str, handler: Handler, name: str):
        super().__init__(path, handler, name, methods={"GET"})


class Post(Path):
    def __init__(self, path: str, handler: Handler, name: str):
        super().__init__(path, handler, name, methods={"POST"})


# ====================
# Static files
# ====================

class Static(Path):
    """Represents a static file route."""

    def __init__(self, path: str, static_directory: str, name: str) -> None:
        self._root = _Path(static_directory).resolve()
        super().__init__(path, self._access, name)

    def _access(self, request: Request, args: dict[str, Optional[str]]) -> Response:
        segments = request.path.strip("/").split("/") if request.path.strip("/") else [""]
        relative_parts = segments[self.detection_range :] if len(segments) > self.detection_range else []
        rel = _Path(*relative_parts) if relative_parts else _Path("")

        try:
            target = (self._root / rel).resolve()
        except Exception:
            raise PathNotFoundError(request.path)

        if not str(target).startswith(str(self._root) + os.sep) and target != self._root:
            raise PathNotFoundError(request.path)

        if target.is_dir():
            target = (target / "index.html").resolve()

        if not target.exists() or not target.is_file():
            raise PathNotFoundError(request.path)

        data = target.read_bytes()
        mime_type, _ = guess_type(str(target))

        headers = [
            ("Content-Type", mime_type or "application/octet-stream"),
            ("Content-Length", str(len(data))),
        ]
        return Response(data, headers=headers)


# ====================
# Pathlib
# ====================

class Router(list[Path]):
    """Collection of Path objects with middleware support."""

    def __init__(self, *paths: Path) -> None:
        for p in paths:
            self._check(p)
        super().__init__(paths)

        self.raise_if_not_exist: bool = True
        self.middlewares: list[Middleware] = []

    def _check(self, arg: object) -> None:
        if not isinstance(arg, Path):
            raise PathlibTypeError

    def add_middleware(self, *middlewares: Middleware) -> None:
        self.middlewares.extend(middlewares)

    def get(self, request_path: str, *, method: str | None = None) -> Path | None:
        segments = request_path.rstrip("/").split("/")
        method_u = method.upper() if method else None

        fallback: Path | None = None

        for p in self:
            detection_path = "/".join(segments[: p.detection_range])
            if detection_path != p.path:
                continue

            # Method-specific Path has priority
            if method_u and p.methods is not None:
                if method_u in p.methods:
                    return p
                continue

            # Method-agnostic Path as fallback
            if fallback is None:
                fallback = p

        if fallback is not None:
            return fallback

        if self.raise_if_not_exist:
            raise PathNotFoundError(request_path)
        return None

    def handle(self, request: Request) -> Response:
        path = self.get(request.path, method=request.method)
        if path is None:
            raise PathNotFoundError(request.path)
        return path.call_handler(request, tuple(self.middlewares))


from __future__ import annotations

from typing import Callable, Iterable
import time

from nexom.app.request import Request
from nexom.app.response import Response, ErrorResponse
from nexom.core.error import PathNotFoundError
from nexom.core.log import AppLogger, AuthLogger

from __app_name__.config import INFO_LOG, WARN_LOG, ERR_LOG, ACES_LOG
from .router import routing


# Logger (global)
logger = AppLogger(
    info=INFO_LOG,
    warn=WARN_LOG,
    error=ERR_LOG,
    access=ACES_LOG,
)


def _ip(environ: dict) -> str:
    xff = environ.get("HTTP_X_FORWARDED_FOR")
    if isinstance(xff, str) and xff.strip():
        return xff.split(",")[0].strip()
    return str(environ.get("REMOTE_ADDR") or "-")


def app(environ: dict, start_response: Callable) -> Iterable[bytes]:
    t0 = time.time()

    req: Request | None = None
    res: Response

    try:
        req = Request(environ)
        path = req.path
        method = req.method

        res = routing.handle(req)

    except PathNotFoundError as e:
        logger.warn(str(e))
        res = ErrorResponse(404, "Not Found")

    except Exception as e:
        logger.error(e)
        res = ErrorResponse(500, "Internal Server Error")

    try:
        dt_ms = int((time.time() - t0) * 1000)
        method = req.method if req else str(environ.get("REQUEST_METHOD") or "-")
        path = (req.path if req else str(environ.get("PATH_INFO") or "")).lstrip("/")
        ip = _ip(environ)
        ua = str(environ.get("HTTP_USER_AGENT") or "-")
        logger.access(f'{ip} "{method} /{path}" {res.status_code} {dt_ms}ms "{ua}"')
    except Exception:
        pass

    start_response(res.status_text, res.headers)
    return [res.body]


from __future__ import annotations

from typing import Iterable, Optional
from importlib import resources
import json

from ..core.object_html_render import HTMLDoc, ObjectHTML

from .http_status_codes import http_status_codes


Header = tuple[str, str]


class Response:
    """
    Represents an HTTP response.
    """

    def __init__(
        self,
        body: str | bytes = b"",
        status: int = 200,
        headers: Iterable[Header] | None = None,
        cookie: str | None = None,
        content_type: str = "text/html",
        charset: str = "utf-8",
        include_charset: bool = False
    ) -> None:
        self.charset: str = charset
        self.include_charset: bool = include_charset

        if isinstance(body, str):
            self.body: bytes = body.encode(charset)
            self.is_text: bool = True
        else:
            self.body = body
            self.is_text = False

        self.status_code: int = status

        from .http_status_codes import http_status_codes
        self.status_text: str = f"{status} {http_status_codes.get(status, '')}".strip()

        ct = content_type
        if include_charset and (ct.startswith("text/") or ct == "application/json"):
            ct = f"{ct}; charset={charset}"

        self.headers: list[Header] = list(headers) if headers else [("Content-Type", ct)]

        if cookie:
            self.headers.append(("Set-Cookie", cookie))

        # ---- auto Content-Length (if not already present) ----
        has_len = any(k.lower() == "content-length" for k, _ in self.headers)
        if not has_len and isinstance(self.body, (bytes, bytearray)):
            self.headers.append(("Content-Length", str(len(self.body))))

    def __iter__(self):
        """
        Allow Response to be returned directly from WSGI apps.
        """
        yield self.body

    def append_header(self, key: str, value: str) -> None:
        self.headers.append((key, value))

class HtmlResponse(Response):
    def __init__(
        self,
        body: str | bytes = b"",
        status: int = 200,
        headers: Iterable[Header] | None = None,
        cookie: str | None = None,
        *,
        charset: str = "utf-8",
        include_charset: bool = True,
    ) -> None:
        super().__init__(
            body=body,
            status=status,
            headers=headers,
            cookie=cookie,
            content_type="text/html",
            charset=charset,
            include_charset=include_charset,
        )


class JsonResponse(Response):
    def __init__(
        self,
        body: Optional[dict] = None,
        status: int = 200,
        headers: Iterable[Header] | None = None,
        cookie: str | None = None,
        *,
        charset: str = "utf-8",
        include_charset: bool = True,
    ) -> None:
        if body is None:
            body = {}

        content_type = "application/json"
        if include_charset:
            content_type = f"{content_type}; charset={charset}"

        jb = json.dumps(body, ensure_ascii=False).encode(charset)

        super().__init__(
            body=jb,
            status=status,
            headers=headers,
            cookie=cookie,
            content_type=content_type,
            charset=charset,
            include_charset=False,
        )

class Redirect(Response):
    """
    HTTP redirect response (302).
    """

    def __init__(
        self,
        location: str,
        headers: Iterable[Header] | None = None,
        cookie: str | None = None,
    ) -> None:
        extra = list(headers) if headers else []
        res_headers = [("Location", location), *extra]

        super().__init__(
            body=b"",
            status=302,
            headers=res_headers,
            cookie=cookie,
        )


class ErrorResponse(Response):
    """
    HTML error response rendered from template.
    """

    def __init__(self, status: int, message: str) -> None:
        html = self._render(status, message)
        
        super().__init__(html, status=status)

    @staticmethod
    def _render(status: int, message: str) -> str:
        status_text = f"{status} {http_status_codes.get(status, '')}".strip()

        template = (
            resources.files("nexom.assets.error_page")
            .joinpath("error.html")
            .read_text(encoding="utf-8")
        )

        ohtml = ObjectHTML(HTMLDoc("error_page", template))

        status_str = f"{status} - {http_status_codes.get(status)}"

        return (
            ohtml.render("error_page", status=status_str, message=message)
        )


from __future__ import annotations

from nexom.app.path import Get, Static, Router
from nexom.templates.auth import AuthPages

from .config import APP_DIR, AUTH_SERVER
from .pages import default, document

routing = Router(
    Get("", default.main, "DefaultPage"),
    Get("doc/", document.main, "DocumentPage"),
    Static("static/", APP_DIR + "/static", "StaticFiles"),

    AuthPages("user/", AUTH_SERVER),
)